"""Swappable models are badly designed in oauth2_provider.

see https://github.com/jazzband/django-oauth-toolkit/issues/634

This custom conf together with the setting
MIGRATION_MODULES = {"oauth2_provider": None}
enables to override all the models without quirks.
"""
import os

from pkg_resources import get_distribution
from django.apps import AppConfig


class Config(AppConfig):
    name = "oauth2_provider"
    verbose_name = "Django OAuth Toolkit"
    path = os.path.join(
        get_distribution("django-oauth-toolkit").location, "oauth2_provider"
    )

    def get_models(self, *args, **kwargs):
        excluded_models = {"Application", "AccessToken", "Grant", "RefreshToken"}
        yield from (
            model
            for model in super().get_models()
            if model.__name__ not in excluded_models
        )



from rest_framework.permissions import BasePermission


def require_scopes(*required_roles):
    class ScopePermission(BasePermission):
        """
        Allows access only to authenticated users with a specific scope
        (or admin rights)
        """

        _required_roles = set(required_roles)

        def has_permission(self, request, view):
            if not request.user or not request.user.is_authenticated:
                return False

            if request.user.is_staff:
                return True

            scopes = getattr(request.user, "scopes", {})
            return self._required_roles.issubset(scopes)

    return ScopePermission



import re

from django.urls import include, path, re_path

from rest_framework import permissions
from rest_framework import routers
from drf_yasg.views import get_schema_view
from drf_yasg import openapi

from . import geographies
from . import policies
from . import service_areas
from . import vehicles
from . import compliances


agency_router = routers.DefaultRouter()
agency_router.register(r"vehicles", vehicles.DeviceViewSet, basename="device")
agency_router.register(r"policies", policies.PolicyViewSet, basename="policy")
agency_router.register(
    r"compliance/snapshot", compliances.ComplianceViewSet, basename="compliance"
)

agency_router.register(
    r"geographies", geographies.GeographyViewSet, basename="geography"
)
# Deprecation pending for "geography"
agency_router.register(r"service_areas", service_areas.AreaViewSet)


def get_url_patterns(prefix):
    # the schema view needs the prefix to generate correct urls
    # => url_patterns cannot be hardcoded here
    prefix = re.split(r"[/]+$", prefix)[0]
    prefix = ("%s/" % prefix) if prefix else ""
    schema_view = get_schema_view(
        openapi.Info(
            title="LADOT agency API",
            default_version="v0.2",
            description="see "
            "https://github.com/CityOfLosAngeles/mobility-data-specification",
        ),
        patterns=[path(prefix, include(agency_router.urls))],
        public=True,
        permission_classes=(permissions.AllowAny,),
    )

    return agency_router.urls + [
        re_path(
            r"^swagger(?P<format>\.json|\.yaml)$",
            schema_view.without_ui(cache_timeout=0),
            name="schema-json",
        ),
        re_path(
            r"^swagger/$",
            schema_view.with_ui("swagger", cache_timeout=0),
            name="schema-swagger-ui",
        ),
    ]



from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Create a service area and 100 devices"

    def add_arguments(self, parser):
        parser.add_argument(
            "--records_count", nargs="?", default=1000, dest="records_count", type=int
        )

    def handle(self, *args, **options):
        # this is a dev-only command
        # mds.factories use factory-boy, a [dev] extra requirement
        from mds import factories

        factories.Area(
            label="District 10", polygons=[factories.Polygon(), factories.Polygon()]
        )
        factories.EventRecord.create_batch(options["records_count"])



import datetime
import importlib
import random

from django.conf import settings
from django.contrib.gis.geos.point import Point

from functools import lru_cache


@lru_cache()
def get_object_from_settings(setting):
    """Return an object.

    ``setting`` should be a full Python path,
    e.g. ``mds.utils.telemetry_is_enabled``.
    """
    module_name, obj_name = setting.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, obj_name)


def telemetry_is_enabled():
    """
    This function is imported with `get_object_from_cache`
    from ENABLE_TELEMETRY_FUNCTION in the settings.
    It may easily be overriden to enable/disable the telemetries
    saving through a dynamic setting.
    """
    return True


is_telemetry_enabled = get_object_from_settings(settings.ENABLE_TELEMETRY_FUNCTION)


def to_mds_timestamp(dt: datetime.datetime) -> int:
    """
    Convert a datetime object into an MDS-compliant timestamp.

    MDS wants the timestamp to be in milliseconds.
    """
    return round(dt.timestamp() * 1000)


def from_mds_timestamp(value: int) -> datetime:
    """Convert an MDS timestamp (in milliseconds) to a datetime object."""
    return datetime.datetime.fromtimestamp(value / 1000, tz=datetime.timezone.utc)


def get_random_point(polygon):
    """Return a random point in the given polygon."""
    (x_min, y_min), (x_max, _), (_, y_max) = polygon.envelope[0][:3]
    while True:
        point = Point(random.uniform(x_min, x_max), random.uniform(y_min, y_max))
        if polygon.contains(point):
            return point



import enum

from django.utils.translation import pgettext_lazy


def choices(enum):
    """Convert an Enum to a Django-style "choices" iterator.
    """
    return [(member.name, member.value) for member in enum]


class DEVICE_STATUS(enum.Enum):
    available = pgettext_lazy("Device status", "Available")
    reserved = pgettext_lazy("Device status", "Reserved")
    unavailable = pgettext_lazy("Device status", "Unavailable")
    removed = pgettext_lazy("Device status", "Removed")
    trip = pgettext_lazy("Device status", "Trip")
    elsewhere = pgettext_lazy("Device status", "Elsewhere")
    inactive = pgettext_lazy("Device status", "Inactive")
    unknown = pgettext_lazy("Device status", "Unknown")


class DEVICE_CATEGORY(enum.Enum):
    bicycle = pgettext_lazy("Device category", "Bicycle")
    scooter = pgettext_lazy("Device category", "Scooter")
    car = pgettext_lazy("Device category", "Car")


class DEVICE_PROPULSION(enum.Enum):
    human = pgettext_lazy("Device propulsion", "Human")
    electric_assist = pgettext_lazy("Device propulsion", "Electric Assist")
    electric = pgettext_lazy("Device propulsion", "Electric")
    combustion = pgettext_lazy("Device propulsion", "Combustion")


class EVENT_TYPE(enum.Enum):
    # The first events are the one listed in the MDS agency API, in order:
    # https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/agency
    register = pgettext_lazy("Event type", "Register")
    service_start = pgettext_lazy("Event type", "Service start")
    service_end = pgettext_lazy("Event type", "Service end")
    provider_drop_off = pgettext_lazy("Event type", "Provider drop off")
    provider_pick_up = pgettext_lazy("Event type", "Provider pick up")
    city_pick_up = pgettext_lazy("Event type", "City pick up")
    reserve = pgettext_lazy("Event type", "Reserve")
    cancel_reservation = pgettext_lazy("Event type", "Cancel reservation")
    trip_start = pgettext_lazy("Event type", "Trip start")
    trip_enter = pgettext_lazy("Event type", "Trip enter")
    trip_leave = pgettext_lazy("Event type", "Trip leave")
    trip_end = pgettext_lazy("Event type", "Trip end")
    deregister = pgettext_lazy("Event type", "Deregister")
    # this last event is in neither of the MDS spec
    telemetry = pgettext_lazy("Event type", "Received telemetry")
    # This may be added in a revision of the agency API spec
    battery_charged = pgettext_lazy("Event type", "Battery charged")


# To make sure we still support the old version of the Agency API
# when we receive events.
class OLD_EVENT_TYPE(enum.Enum):
    rebalance_drop_off = pgettext_lazy("Event type", "Rebalance drop off")
    maintenance_drop_off = pgettext_lazy("Event type", "Maintenance drop off")
    low_battery = pgettext_lazy("Event type", "Low battery")
    maintenance = pgettext_lazy("Event type", "Maintenance")
    rebalance_pick_up = pgettext_lazy("Event type", "Rebalance pick up")
    maintenance_pick_up = pgettext_lazy("Event type", "Maintenance pick up")


# event_type_reason(s) listed in the provider API but not in the agency API
# https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/provider
class PROVIDER_EVENT_TYPE_REASON(enum.Enum):
    # available
    service_start = pgettext_lazy("Event type", "Service start")
    user_drop_off = pgettext_lazy("Event type", "User drop off")
    rebalance_drop_off = pgettext_lazy("Event type", "Rebalance drop off")
    maintenance_drop_off = pgettext_lazy("Event type", "Maintenance drop off")
    agency_drop_off = pgettext_lazy("Event type", "Agency drop off")
    # reserved
    user_pick_up = pgettext_lazy("Event type", "User pick up")
    # unavailable
    maintenance = pgettext_lazy("Event type", "Maintenance")
    low_battery = pgettext_lazy("Event type", "Low battery")
    # removed
    service_end = pgettext_lazy("Event type", "Service end")
    rebalance_pick_up = pgettext_lazy("Event type", "Rebalance pick up")
    maintenance_pick_up = pgettext_lazy("Event type", "Maintenance pick up")
    agency_pick_up = pgettext_lazy("Event type", "Agency pick up")
    # this last event is in neither of the MDS spec
    telemetry = pgettext_lazy("Event type", "Received telemetry")
    # This may be added in a revision of the agency API spec
    battery_charged = pgettext_lazy("Event type", "Battery charged")


# see: https://github.com/CityOfLosAngeles/mobility-data-specification/tree/dev/agency
class EVENT_TYPE_REASON(enum.Enum):
    low_battery = pgettext_lazy("Event type reason", "Low battery")
    maintenance = pgettext_lazy("Event type reason", "Maintenance")
    compliance = pgettext_lazy("Event type reason", "Compliance")
    off_hours = pgettext_lazy("Event type reason", "Off hours")
    rebalance = pgettext_lazy("Event type reason", "Rebalance")
    charge = pgettext_lazy("Event type reason", "Charge")
    missing = pgettext_lazy("Event type reason", "Missing")
    decommissioned = pgettext_lazy("Event type reason", "Decommissioned")


class EVENT_SOURCE(enum.Enum):
    agency_api = pgettext_lazy("Event source", "Agency API")
    provider_api = pgettext_lazy("Event source", "Provider API")


class AREA_TYPE(enum.Enum):
    unrestricted = pgettext_lazy("Area type", "Unrestricted")
    restricted = pgettext_lazy("Area type", "Restricted")
    preferred_pick_up = pgettext_lazy("Area type", "Preferred pick up")
    preferred_drop_off = pgettext_lazy("Area type", "Preferred drop off")


EVENT_TYPE_TO_DEVICE_STATUS = {
    EVENT_TYPE.register.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.service_start.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.service_end.name: DEVICE_STATUS.unavailable.name,
    EVENT_TYPE.provider_drop_off.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.provider_pick_up.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.city_pick_up.name: DEVICE_STATUS.removed.name,
    EVENT_TYPE.reserve.name: DEVICE_STATUS.reserved.name,
    EVENT_TYPE.cancel_reservation.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.trip_start.name: DEVICE_STATUS.trip.name,
    EVENT_TYPE.trip_enter.name: DEVICE_STATUS.trip.name,
    EVENT_TYPE.trip_leave.name: DEVICE_STATUS.elsewhere.name,
    EVENT_TYPE.trip_end.name: DEVICE_STATUS.available.name,
    EVENT_TYPE.deregister.name: DEVICE_STATUS.inactive.name,
    # Other events
    EVENT_TYPE.telemetry.name: DEVICE_STATUS.unknown.name,
    EVENT_TYPE.battery_charged.name: DEVICE_STATUS.available.name,
    # Old events
    OLD_EVENT_TYPE.rebalance_drop_off.name: DEVICE_STATUS.available.name,
    OLD_EVENT_TYPE.maintenance_drop_off.name: DEVICE_STATUS.available.name,
    OLD_EVENT_TYPE.low_battery.name: DEVICE_STATUS.unavailable.name,
    OLD_EVENT_TYPE.maintenance.name: DEVICE_STATUS.unavailable.name,
    OLD_EVENT_TYPE.rebalance_pick_up.name: DEVICE_STATUS.removed.name,
    OLD_EVENT_TYPE.maintenance_pick_up.name: DEVICE_STATUS.removed.name,
}


class POLICY_RULE_TYPES(enum.Enum):
    count = pgettext_lazy("Policy type", "Fleet size")
    time = pgettext_lazy("Policy type", "Time limit")
    speed = pgettext_lazy("Policy type", "Speed limit")
    user = pgettext_lazy("Policy type", "Information for users")
    geofence = pgettext_lazy("Policy type", "Geofencing")



# Generated by Django 2.2.7 on 2019-12-18 10:26

import uuid

import django.contrib.gis.db.models.fields
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
import django.contrib.postgres.functions
import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone

import mds.models
import rest_framework.utils.encoders


class Migration(migrations.Migration):

    replaces = [
        ("mds", "0001_initial"),
        ("mds", "0002_area_color"),
        ("mds", "0003_record_timestamp"),
        ("mds", "0004_eventrecord_point"),
        ("mds", "0005_auto_20190122_1300"),
        ("mds", "0006_provider_device_category"),
        ("mds", "0007_provider_api"),
        ("mds", "0008_provider_authentication"),
        ("mds", "0009_provider_agency_api_authentication"),
        ("mds", "0010_device_dn_battery_pct"),
        ("mds", "0011_added_index_to_polygon_and_area_models"),
        ("mds", "0012_update_to_mds_0_3_0"),
        ("mds", "0013_provider_last_start_time_polled"),
        ("mds", "0014_non_null_labels"),
        ("mds", "0015_null_dn_status"),
        ("mds", "0016_agency_api_configuration"),
        ("mds", "0017_change_bike_to_bicycle_in_choices_field"),
        ("mds", "0018_allow_blank_provider_base_api_url"),
        ("mds", "0019_delete_creation_deletion_date_fields_in_area_and_polygon"),
        ("mds", "0020_fields_consistency"),
        ("mds", "0021_rename_event_source"),
        ("mds", "0022_provider_colors"),
        ("mds", "0023_relabel_api_fields"),
        ("mds", "0024_provider_device_category_deprecation"),
        ("mds", "0025_remove_provider_device_category"),
        ("mds", "0026_alter_polygon_to_multipolygon"),
        ("mds", "0027_eventrecord_first_saved_at"),
        ("mds", "0028_pre_provider_operator"),
        ("mds", "0029_post_provider_operator_fill"),
        ("mds", "0030_post_provider_operator_default"),
        ("mds", "0031_post_eventrecord_saved_at_index"),
        ("mds", "0032_pre_device_saved_at"),
        ("mds", "0033_post_fill_device_saved_at"),
        ("mds", "0034_post_default_device_saved_at"),
        ("mds", "0035_pre_add_fields"),
        ("mds", "0036_pre_save_provider_cursors"),
        ("mds", "0037_remove_provider_last_start_time_polled"),
        ("mds", "0038_eventrecord_event_type_reason"),
        ("mds", "0039_alter_field_event_type_event_record"),
        ("mds", "0040_change_event_record_event_type_options"),
        ("mds", "0041_map_old_to_new_event_records"),
        ("mds", "0042_revert_0039_event_type"),
        ("mds", "0043_policy"),
        ("mds", "0044_add_compliance_list"),
        ("mds", "0045_add_fixed_price_on_policy_model"),
        ("mds", "0046_compliance_uniqueness"),
        ("mds", "0047_add_daily_penalty_20190924_0933"),
        ("mds", "0048_add_compliance_lag"),
        ("mds", "0049_pre_add_policy_config"),
        ("mds", "0050_pre_migrate_to_policy_config"),
        ("mds", "0051_fill_compliance_lag"),
        ("mds", "0052_required_compliance_lag"),
        ("mds", "0053_post_remove_policy_custom_fields"),
        ("mds", "0054_pre_compliance_saved_at_nullable"),
        ("mds", "0055_post_fill_compliance_saved_at"),
        ("mds", "0056_post_compliance_saved_at_not_nullable"),
        ("mds", "0057_pre_policy_geographies_add"),
        ("mds", "0058_post_policy_geograpies_fill"),
        ("mds", "0059_post_policy_config_encoder"),
    ]

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Polygon",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                (
                    "label",
                    mds.models.UnboundedCharField(
                        blank=True, db_index=True, default=""
                    ),
                ),
                (
                    "geom",
                    django.contrib.gis.db.models.fields.MultiPolygonField(srid=4326),
                ),
                (
                    "properties",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict, encoder=rest_framework.utils.encoders.JSONEncoder
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Provider",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("name", mds.models.UnboundedCharField(blank=True, default="")),
                (
                    "logo_b64",
                    mds.models.UnboundedCharField(blank=True, default=None, null=True),
                ),
                ("operator", models.BooleanField(default=True)),
                (
                    "base_api_url",
                    mds.models.UnboundedCharField(
                        blank=True, default="", verbose_name="Base Provider API URL"
                    ),
                ),
                (
                    "oauth2_url",
                    mds.models.UnboundedCharField(
                        blank=True, default="", verbose_name="OAuth2 URL (if different)"
                    ),
                ),
                (
                    "api_authentication",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        default=mds.models.provider_api_authentication_default,
                        verbose_name="Provider API Authentication",
                    ),
                ),
                (
                    "api_configuration",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        default=mds.models.provider_api_configuration_default,
                        verbose_name="Provider API Configuration",
                    ),
                ),
                (
                    "last_event_time_polled",
                    models.DateTimeField(
                        blank=True,
                        null=True,
                        verbose_name="Last event time polled (start_time)",
                    ),
                ),
                (
                    "last_recorded_polled",
                    models.DateTimeField(
                        blank=True,
                        null=True,
                        verbose_name="Last recorded polled (start_recorded)",
                    ),
                ),
                (
                    "last_skip_polled",
                    models.PositiveIntegerField(
                        blank=True,
                        null=True,
                        verbose_name="Last skip polled (total_events)",
                    ),
                ),
                (
                    "agency_api_authentication",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        default=mds.models.agency_api_authentication_default,
                        verbose_name="Agency API Authentication",
                    ),
                ),
                (
                    "agency_api_configuration",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        default=mds.models.agency_api_configuration_default,
                        verbose_name="Agency API Configuration",
                    ),
                ),
                (
                    "colors",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        default=dict,
                        validators=[mds.models.colors_validator],
                        verbose_name="colors for distinguishing provider icons",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Policy",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("name", mds.models.UnboundedCharField()),
                ("description", mds.models.UnboundedCharField(blank=True, default="")),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField(blank=True, null=True)),
                ("published_date", models.DateTimeField(blank=True, null=True)),
                (
                    "rules",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=list,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                (
                    "config",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                (
                    "geographies",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "prev_policies",
                    models.ManyToManyField(
                        blank=True,
                        related_name="_policy_prev_policies_+",
                        to="mds.Policy",
                    ),
                ),
                ("providers", models.ManyToManyField(blank=True, to="mds.Provider")),
            ],
            options={"verbose_name_plural": "policies"},
        ),
        migrations.CreateModel(
            name="Device",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                (
                    "registration_date",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("identification_number", mds.models.UnboundedCharField()),
                (
                    "category",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("bicycle", "Bicycle"),
                            ("scooter", "Scooter"),
                            ("car", "Car"),
                        ]
                    ),
                ),
                ("model", mds.models.UnboundedCharField(blank=True, default="")),
                (
                    "propulsion",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=mds.models.UnboundedCharField(
                            choices=[
                                ("human", "Human"),
                                ("electric_assist", "Electric Assist"),
                                ("electric", "Electric"),
                                ("combustion", "Combustion"),
                            ]
                        ),
                        size=None,
                    ),
                ),
                ("year_manufactured", models.IntegerField(blank=True, null=True)),
                ("manufacturer", mds.models.UnboundedCharField(blank=True, default="")),
                ("saved_at", models.DateTimeField(auto_now=True, db_index=True)),
                ("dn_battery_pct", models.FloatField(blank=True, null=True)),
                (
                    "dn_gps_point",
                    django.contrib.gis.db.models.fields.PointField(
                        blank=True, null=True, srid=4326
                    ),
                ),
                ("dn_gps_timestamp", models.DateTimeField(blank=True, null=True)),
                (
                    "dn_status",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("available", "Available"),
                            ("reserved", "Reserved"),
                            ("unavailable", "Unavailable"),
                            ("removed", "Removed"),
                            ("trip", "Trip"),
                            ("elsewhere", "Elsewhere"),
                            ("inactive", "Inactive"),
                            ("unknown", "Unknown"),
                        ],
                        default="unknown",
                    ),
                ),
                (
                    "provider",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="devices",
                        to="mds.Provider",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Area",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                (
                    "label",
                    mds.models.UnboundedCharField(
                        blank=True, db_index=True, default=""
                    ),
                ),
                (
                    "color",
                    mds.models.UnboundedCharField(
                        default="#FFFFFF", help_text="hexa representation"
                    ),
                ),
                (
                    "area_type",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("unrestricted", "Unrestricted"),
                            ("restricted", "Restricted"),
                            ("preferred_pick_up", "Preferred pick up"),
                            ("preferred_drop_off", "Preferred drop off"),
                        ],
                        default="unrestricted",
                    ),
                ),
                (
                    "polygons",
                    models.ManyToManyField(
                        blank=True, related_name="areas", to="mds.Polygon"
                    ),
                ),
                (
                    "providers",
                    models.ManyToManyField(
                        blank=True, related_name="areas", to="mds.Provider"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EventRecord",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(db_index=True)),
                (
                    "point",
                    django.contrib.gis.db.models.fields.PointField(
                        blank=True, null=True, srid=4326
                    ),
                ),
                (
                    "saved_at",
                    models.DateTimeField(
                        db_index=True,
                        default=django.contrib.postgres.functions.TransactionNow,
                    ),
                ),
                ("first_saved_at", models.DateTimeField(blank=True, null=True)),
                (
                    "source",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("agency_api", "Agency API"),
                            ("provider_api", "Provider API"),
                        ],
                        default="agency_api",
                    ),
                ),
                (
                    "event_type",
                    mds.models.UnboundedCharField(
                        choices=[
                            ("register", "Register"),
                            ("service_start", "Service start"),
                            ("service_end", "Service end"),
                            ("provider_drop_off", "Provider drop off"),
                            ("provider_pick_up", "Provider pick up"),
                            ("city_pick_up", "City pick up"),
                            ("reserve", "Reserve"),
                            ("cancel_reservation", "Cancel reservation"),
                            ("trip_start", "Trip start"),
                            ("trip_enter", "Trip enter"),
                            ("trip_leave", "Trip leave"),
                            ("trip_end", "Trip end"),
                            ("deregister", "Deregister"),
                            ("telemetry", "Received telemetry"),
                            ("battery_charged", "Battery charged"),
                            ("rebalance_drop_off", "Rebalance drop off"),
                            ("maintenance_drop_off", "Maintenance drop off"),
                            ("low_battery", "Low battery"),
                            ("maintenance", "Maintenance"),
                            ("rebalance_pick_up", "Rebalance pick up"),
                            ("maintenance_pick_up", "Maintenance pick up"),
                        ]
                    ),
                ),
                (
                    "event_type_reason",
                    mds.models.UnboundedCharField(
                        blank=True,
                        choices=[
                            ("low_battery", "Low battery"),
                            ("maintenance", "Maintenance"),
                            ("compliance", "Compliance"),
                            ("off_hours", "Off hours"),
                            ("rebalance", "Rebalance"),
                            ("charge", "Charge"),
                            ("missing", "Missing"),
                            ("decommissioned", "Decommissioned"),
                        ],
                        null=True,
                    ),
                ),
                (
                    "properties",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        default=dict, encoder=rest_framework.utils.encoders.JSONEncoder
                    ),
                ),
                (
                    "device",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="event_records",
                        to="mds.Device",
                    ),
                ),
            ],
            options={"unique_together": {("device", "timestamp")}},
        ),
        migrations.CreateModel(
            name="Compliance",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("rule", models.UUIDField()),
                ("geography", models.UUIDField()),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField(blank=True, null=True)),
                ("lag", models.DurationField()),
                ("saved_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "policy",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="compliances",
                        to="mds.Policy",
                    ),
                ),
                (
                    "vehicle",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="compliances",
                        to="mds.Device",
                    ),
                ),
            ],
            options={
                "unique_together": {
                    ("policy", "rule", "geography", "vehicle", "start_date", "lag")
                },
            },
        ),
    ]
